@@ -30,6 +30,34 @@ int debug = 0;
 char * mailCommand = DEFAULT_MAIL_COMMAND;
 time_t nowSecs = 0;
 
+/* converts ' to '\''
+ * returns pointer to a newly allocated string. */
+char *escapeSingleQuotes(const char *src)
+{
+	int c = 0;
+	int i;
+	char *newstr, *s;
+
+	for (i = 0; i < strlen(src); i++) {
+		if (src[i] == '\'') c++;
+	}
+	newstr = malloc(strlen(src) + (c * 3) + 1);
+
+	s = newstr;
+	for (i = 0; i < strlen(src); i++) {
+		if (src[i] == '\'') {
+			strcpy(s, "'\\''");
+			s += 4;
+		} else {
+			*s = src[i];
+			s++;
+		}
+	}
+	*s = '\0';
+
+	return newstr;
+}
+
 static logState * findState(const char * fn, logState ** statesPtr, 
 			    int * numStatesPtr) {
     int i;
@@ -65,8 +93,9 @@ static logState * findState(const char * fn, logState ** statesPtr,
 
 static int runScript(char * logfn, char * script) {
     int fd;
-    char *filespec, *cmd;
+    char *filespec, *cmd, *escapedlogfn;
     int rc;
+    char buf[256];
 
     if (debug) {
 	message(MESS_DEBUG, "running script with arg %s: \"%s\"\n", 
@@ -74,9 +103,8 @@ static int runScript(char * logfn, char * script) {
 	return 0;
     }
 
-    if (asprintf(&filespec,
-	"%s/logrotate.XXXXXX", getenv("TMPDIR") ?: "/tmp") < 0)
-	filespec = NULL;
+    filespec = buf;
+    snprintf(buf, sizeof(buf), "%s/logrotate.XXXXXX", getenv("TMPDIR") ?: "/tmp");
     fd = -1;
     if (!filespec || (fd = mkstemp(filespec)) < 0 || fchmod(fd, 0700)) {
 	message(MESS_DEBUG, "error creating %s: %s\n", filespec,
@@ -85,8 +113,6 @@ static int runScript(char * logfn, char * script) {
 	    close(fd);
 	    unlink(filespec);
 	}
-	if (filespec)
-	    free(filespec);
 	return -1;
     }
 
@@ -95,18 +121,18 @@ static int runScript(char * logfn, char * script) {
 	message(MESS_DEBUG, "error writing %s\n", filespec);
 	close(fd);
 	unlink(filespec);
-	free(filespec);
 	return -1;
     }
 
     close(fd);
 
-    cmd = alloca(strlen(filespec) + strlen(logfn) + 20);
-    sprintf(cmd, "/bin/sh %s '%s'", filespec, logfn);
+    escapedlogfn = escapeSingleQuotes(logfn);
+    cmd = alloca(strlen(filespec) + strlen(escapedlogfn) + 20);
+    sprintf(cmd, "/bin/sh %s '%s'", filespec, escapedlogfn);
+    free(escapedlogfn);
     rc = system(cmd);
 
     unlink(filespec);
-    free(filespec);
 
     return rc;
 }
@@ -294,6 +320,7 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
     char * baseName;
     char * dirName;
     char * firstRotated;
+    size_t alloc_size;
     int rotateCount = log->rotateCount ? log->rotateCount : 1;
     int logStart = (log->logStart == -1) ? 1 : log->logStart;
 
@@ -312,17 +339,28 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
     state->lastRotated = now;
     
     if (log->oldDir)
-        dirName = strdup(log->oldDir);
+      {
+	if(log->oldDir[0] != '/')
+	  {
+	    char *ld = ourDirName(log->files[logNum]);
+	    dirName = malloc(strlen(ld) + strlen(log->oldDir) + 2);
+	    sprintf(dirName, "%s/%s", ld, log->oldDir);
+	    free(ld);
+	  }
+	else
+	  dirName = strdup(log->oldDir);
+      }
     else
         dirName = ourDirName(log->files[logNum]);
     baseName = strdup(ourBaseName(log->files[logNum]));
+
+    alloc_size = strlen(dirName) + strlen(baseName) + 
+                 strlen(log->files[logNum]) + strlen(fileext) +
+                 strlen(compext) + 10;
     
-    oldName = alloca(strlen(dirName) + strlen(baseName) + 
-                     strlen(log->files[logNum]) + 10);
-    newName = alloca(strlen(dirName) + strlen(baseName) + 
-                     strlen(log->files[logNum]) + 10);
-    disposeName = alloca(strlen(dirName) + strlen(baseName) + 
-                         strlen(log->files[logNum]) + 10);
+    oldName = alloca(alloc_size);
+    newName = alloca(alloc_size);
+    disposeName = alloca(alloc_size);
     
     if (log->extension &&
 	strncmp(&baseName[strlen(baseName)-strlen(log->extension)],
@@ -347,13 +385,16 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
             message(MESS_DEBUG, "previous log %s does not exist\n",
 		    oldName);
         } else {
+            char * escapedName;
             char * command;
 	    
+            escapedName = escapeSingleQuotes(oldName);
             command = alloca(strlen(log->compress_prog) +
                              strlen(log->compress_options) +
-                             strlen(oldName) + 10);
+                             strlen(escapedName) + 10);
             sprintf(command, "%s %s '%s'", log->compress_prog,
-                    log->compress_options, oldName);
+                    log->compress_options, escapedName);
+            free(escapedName);
             message(MESS_DEBUG, "compressing previous log with: %s\n",
 		    command);
             if (!debug && system(command)) {
@@ -375,7 +416,7 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
     sprintf(firstRotated, "%s/%s.%d%s%s", dirName, baseName,
             logStart, fileext, compext);
     
-    for (i = rotateCount + logStart; (i >= 0) && !hasErrors; i--) {
+    for (i = rotateCount + logStart - 1; (i >= 0) && !hasErrors; i--) {
         tmp = newName;
         newName = oldName;
         oldName = tmp;
@@ -458,7 +499,7 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
                 createMode = log->createMode;
 	    
             message(MESS_DEBUG, "creating new log mode = 0%o uid = %d "
-                    "gid = %d\n", createMode, createUid, createGid);
+                    "gid = %d\n", (unsigned int)createMode, (int)createUid, (int)createGid);
 	    
             if (!debug) {
                 fd = open(log->files[logNum], O_CREAT | O_RDWR, createMode);
@@ -507,13 +548,16 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
         if (!hasErrors && 
 	    (log->flags & LOG_FLAG_COMPRESS) &&
 	    !(log->flags & LOG_FLAG_DELAYCOMPRESS)) {
+            char * escapedName;
             char * command;
 	    
+            escapedName = escapeSingleQuotes(finalName);
             command = alloca(strlen(log->compress_prog) +
                              strlen(log->compress_options) +
-                             strlen(oldName) + 10);
+                             strlen(escapedName) + 10);
             sprintf(command, "%s %s '%s'", log->compress_prog,
-                    log->compress_options, finalName);
+                    log->compress_options, escapedName);
+            free(escapedName);
             message(MESS_DEBUG, "compressing new log with: %s\n", command);
             if (!debug && system(command)) {
                 fprintf(errorFile, "failed to compress log %s\n", 
@@ -532,24 +576,34 @@ int rotateSingleLog(logInfo * log, int logNum, logState ** statesPtr,
                 mailFilename = disposeName;
 	    
             if (mailFilename) {
-                command = alloca(strlen(mailFilename) + 100 + 
-                                 strlen(log->uncompress_prog));
-		
+		char * escapedMailFilename = escapeSingleQuotes(mailFilename);
                 if ((log->flags & LOG_FLAG_COMPRESS) &&
 		    !(log->flags & LOG_FLAG_DELAYCOMPRESS) &&
-		    (log->flags & LOG_FLAG_MAILFIRST))
-                    sprintf(command, "%s < %s | %s '%s' %s", 
-			    log->uncompress_prog, mailFilename, mailCommand,
-			    log->files[logNum],
-			    log->logAddress);
-                else
-                    sprintf(command, "%s '%s' %s < %s", mailCommand, 
-			    mailFilename, log->logAddress, mailFilename);
+		    (log->flags & LOG_FLAG_MAILFIRST)) {
+                    char * escapedFN = escapeSingleQuotes(log->files[logNum]);
+                    command = alloca(strlen(log->uncompress_prog) +
+				     strlen(escapedMailFilename) +
+				     strlen(mailCommand) +
+				     strlen(escapedFN) +
+				     strlen(log->logAddress) + 20);
+                    sprintf(command, "%s < '%s' | %s '%s' %s", 
+			    log->uncompress_prog, escapedMailFilename,
+			    mailCommand, escapedFN, log->logAddress);
+		    free(escapedFN);
+		} else {
+                    command = alloca(strlen(mailCommand) +
+				     strlen(escapedMailFilename) +
+				     strlen(log->logAddress) +
+				     strlen(escapedMailFilename) + 20);
+                    sprintf(command, "%s '%s' %s < '%s'", mailCommand, 
+			    escapedMailFilename, log->logAddress, escapedMailFilename);
+                }
+		free(escapedMailFilename);
 		
                 message(MESS_DEBUG, "executing: \"%s\"\n", command);
 		
                 if (!debug && system(command)) {
-                    sprintf(newName, "%s.%d", log->files[logNum], getpid());
+                    sprintf(newName, "%s.%d", log->files[logNum], (unsigned int)getpid());
                     fprintf(errorFile, "Failed to mail %s to %s!\n",
                             mailFilename, log->logAddress);
 		    
@@ -627,6 +681,20 @@ int rotateLogSet(logInfo * log, logState ** statesPtr, int * numStatesPtr,
             numRotated++;
     }
     
+    if (log->first) {
+        if (!numRotated) {
+            message(MESS_DEBUG, "not running first action script, "
+                    "since no logs will be rotated\n");
+        } else {
+            message(MESS_DEBUG, "running first action script\n");
+            if (runScript(log->pattern, log->first)) {
+                fprintf(stderr, "error running first action script "
+                        "for %s\n", log->pattern);
+                hasErrors = 1;
+            }
+	}
+    }
+
     if (log->pre && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) {
         if (!numRotated) {
             message(MESS_DEBUG, "not running shared prerotate script, "
@@ -659,6 +727,20 @@ int rotateLogSet(logInfo * log, logState ** statesPtr, int * numStatesPtr,
 	}
     }
     
+    if (log->last) {
+        if (!numRotated) {
+            message(MESS_DEBUG, "not running last action script, "
+                    "since no logs will be rotated\n");
+        } else {
+            message(MESS_DEBUG, "running last action script\n");
+            if (runScript(log->pattern, log->last)) {
+                fprintf(stderr, "error running last action script "
+                        "for %s\n", log->pattern);
+                hasErrors = 1;
+            }
+	}
+    }
+    
     return hasErrors;
 }
 
@@ -831,6 +913,7 @@ int main(int argc, const char ** argv) {
 			  /* threshHold */ 1024 * 1024, 0,
 			  /* log start */ -1,
 			  /* pre, post */ NULL, NULL,
+			  /* first, last */ NULL, NULL,
 			  /* logAddress */ NULL, 
 			  /* extension */ NULL, 
 			  /* compression */ COMPRESS_COMMAND,
@@ -885,7 +968,7 @@ int main(int argc, const char ** argv) {
 	return 2;
     }
 
-    files = poptGetArgs(optCon);
+    files = poptGetArgs((poptContext) optCon);
     if (!files) {
 	fprintf(stderr, "logrotate " VERSION 
 		" - Copyright (C) 1995-2001 Red Hat, Inc.\n");
